import tempfile
from io import StringIO
from pathlib import Path
from textwrap import dedent

import cwl_utils.parser_v1_1 as parser
from cwltool.load_tool import fetch_document, resolve_and_validate_document
from WDL import parse_document

from acclimatise.yaml import yaml


def validate_cwl(cwl: str):
    parsed = yaml.load(cwl)
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        tmpfile = tmpdir / parsed["id"]
        tmpfile.write_text(cwl)
        loading_context, workflowobj, uri = fetch_document(str(tmpfile))
        resolve_and_validate_document(loading_context, workflowobj, uri)


def validate_wdl(wdl: str):
    return parse_document(wdl)



# from acclimatise.flag_parser.elements import cli_id, any_flag, long_flag, short_flag, flag_with
from acclimatise.usage_parser.model import UsageElement

from .elements import *


def parse_usage(cmd, text):
    toks = usage.searchString(text)
    if not toks:
        # If we had no results, return an empty command
        return Command(command=cmd, positional=[], named=[])

    toks = toks[0]

    positional = [tok for tok in toks if isinstance(tok, UsageElement)]
    flags = [tok for tok in toks if isinstance(tok, Flag)]

    # Remove an "options" argument which is just a proxy for other flags
    positional = [pos for pos in positional if pos.text != "options"]

    # The usage often starts with a re-iteration of the command name itself. Remove this if present
    truncate = 0
    for i, arg in enumerate(cmd):
        pos = positional[i]
        if arg == pos.text:
            truncate = i + 1
        else:
            break
    positional = positional[truncate:]

    if not any([tok for tok in positional if tok.variable]):
        # If the usage didn't explicitly mark anything as a variable using < > brackets, we have to assume that
        # everything other than flags are positional elements
        for element in positional:
            element.variable = True

    return Command(
        command=cmd,
        positional=[
            Positional(description="", position=i, name=el.text, optional=el.optional)
            for i, el in enumerate(positional)
        ],
        named=flags,
    )



"""
Functions for generating WDL from the CLI data model
"""
from pathlib import Path
from typing import Generator, Iterable, List

from inflection import camelize
from wdlgen import ArrayType, Input, PrimitiveType, Task, WdlType

from acclimatise import cli_types, model
from acclimatise.converter import WrapperGenerator
from acclimatise.model import Command


def flag_to_command_input(flag: model.CliArgument) -> Task.Command.CommandInput:
    args = dict(name=flag.name_to_camel())

    if isinstance(flag, model.Flag):
        args.update(dict(optional=True,))
        if isinstance(flag.args, model.EmptyFlagArg):
            args.update(dict(true=flag.longest_synonym, false=""))
        else:
            args.update(dict(prefix=flag.longest_synonym,))
    elif isinstance(flag, model.Positional):
        args.update(dict(optional=False, position=flag.position))

    return Task.Command.CommandInput(**args)


def lowest_common_type(types: Iterable[cli_types.CliType]):
    type_set = {type(t) for t in types}

    if len(type_set) == 1:
        # If there is only one type, use it
        return type_set.pop()

    if (
        len(type_set) == 2
        and cli_types.CliInteger in type_set
        and cli_types.CliFloat in type_set
    ):
        # If they're all numeric, they can be represented as floats
        return cli_types.CliFloat()

    if {
        cli_types.CliDir,
        cli_types.CliDict,
        cli_types.CliFile,
        cli_types.CliTuple,
        cli_types.CliList,
    } & type_set:
        # These complex types cannot be represented in a simpler way
        raise Exception(
            "There is no common type between {}".format(", ".join(type_set))
        )

    else:
        # Most of the time, strings can be used to represent primitive types
        return cli_types.CliString()


class WdlGenerator(WrapperGenerator):
    case = "snake"

    @classmethod
    def format(cls) -> str:
        return "wdl"

    @classmethod
    def type_to_wdl(cls, typ: cli_types.CliType, optional: bool = False) -> WdlType:
        if isinstance(typ, cli_types.CliString):
            return WdlType(PrimitiveType(PrimitiveType.kString), optional=optional)
        elif isinstance(typ, cli_types.CliFloat):
            return WdlType(PrimitiveType(PrimitiveType.kFloat), optional=optional)
        elif isinstance(typ, cli_types.CliBoolean):
            return WdlType(PrimitiveType(PrimitiveType.kBoolean), optional=optional)
        elif isinstance(typ, cli_types.CliInteger):
            return WdlType(PrimitiveType(PrimitiveType.kInt), optional=optional)
        elif isinstance(typ, cli_types.CliFile):
            return WdlType(PrimitiveType(PrimitiveType.kFile), optional=optional)
        elif isinstance(typ, cli_types.CliDir):
            return WdlType(PrimitiveType(PrimitiveType.kDirectory), optional=optional)
        elif isinstance(typ, cli_types.CliTuple):
            if typ.homogenous:
                return WdlType(
                    ArrayType(
                        cls.type_to_wdl(typ.values[0]), requires_multiple=not optional
                    )
                )
            else:
                return WdlType(
                    ArrayType(
                        cls.type_to_wdl(lowest_common_type(typ.values)),
                        requires_multiple=not optional,
                    )
                )
        elif isinstance(typ, cli_types.CliList):
            return WdlType(
                ArrayType(cls.type_to_wdl(typ.value), requires_multiple=not optional)
            )
        elif isinstance(typ, cli_types.CliEnum):
            return WdlType(PrimitiveType(PrimitiveType.kString), optional=optional)
        else:
            return WdlType(PrimitiveType(PrimitiveType.kString), optional=optional)

    def generate_wrapper(self, cmd: Command) -> str:
        name = cmd.as_filename
        task_name = camelize(name)

        inputs = [
            Input(
                data_type=self.type_to_wdl(pos.get_type(), optional=False),
                name=pos.name_to_camel(),
            )
            for pos in cmd.named
        ]
        if not self.ignore_positionals:
            inputs += [
                Input(
                    data_type=self.type_to_wdl(pos.get_type(), optional=True),
                    name=pos.name_to_camel(),
                )
                for pos in cmd.positional
            ]

        tool = Task(
            name=task_name,
            command=Task.Command(
                command=" ".join(cmd.command),
                inputs=[flag_to_command_input(pos) for pos in cmd.positional],
                arguments=[flag_to_command_input(named) for named in cmd.named],
            ),
            version="1.0",
            inputs=inputs,
        )

        return tool.get_string()

    def generate_tree(self, cmd: Command, out_dir: Path) -> Generator[Path, None, None]:
        out_dir = Path(out_dir)
        for command in cmd.command_tree():
            name = command.as_filename
            path = (out_dir / name).with_suffix(".wdl")
            wrapper = self.generate_wrapper(command)
            path.write_text(wrapper, encoding="utf-8")
            yield path



from pathlib import Path
from typing import List

from dataclasses import dataclass

from acclimatise.converter import WrapperGenerator
from acclimatise.model import Command
from acclimatise.yaml import yaml


@dataclass
class YmlGenerator(WrapperGenerator):
    """
    Internal YML format
    """

    @classmethod
    def format(cls) -> str:
        return "yml"

    def generate_wrapper(self, cmd: Command) -> str:
        return yaml.dump(cmd)

    def generate_tree(self, cmd: Command, out_dir: Path) -> List[Path]:
        path = (out_dir / cmd.as_filename).with_suffix(".yml")
        with path.open("w") as fp:
            yaml.dump(cmd, fp)
        return [path]



import shutil
import tempfile
import traceback
from pathlib import Path

import pytest
from click.testing import CliRunner

from acclimatise.cli import main

from .util import validate_cwl, validate_wdl


@pytest.fixture()
def runner():
    return CliRunner()


def cli_worked(result):
    if result.exit_code == 0:
        return True
    else:
        traceback.print_exception(*result.exc_info)
        assert result.exit_code == 0


def test_pipe_wdl(runner, htseq_help):
    result = runner.invoke(
        main, ["pipe", "htseq-count", "--format", "wdl"], input=htseq_help
    )
    cli_worked(result)
    validate_wdl(result.output)


def test_pipe_cwl(runner, htseq_help):
    result = runner.invoke(
        main, ["pipe", "htseq-count", "--format", "cwl"], input=htseq_help
    )
    cli_worked(result)
    validate_cwl(result.output)


@pytest.mark.skipif(
    not shutil.which("htseq-count"), reason="htseq-count is not installed"
)
def test_explore_htseq(runner):
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(main, ["explore", "htseq-count", "--out-dir", tempdir])
        cli_worked(result)
        assert len(list(Path(tempdir).iterdir())) == 3


@pytest.mark.skipif(not shutil.which("samtools"), reason="samtools is not installed")
def test_explore_samtools(runner):
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            main, ["explore", "samtools", "--help-flag", "--help", "--out-dir", tempdir]
        )
        cli_worked(result)
        assert len(list(Path(tempdir).iterdir())) > 20


@pytest.mark.skipif(not shutil.which("samtools"), reason="samtools is not installed")
def test_explore_samtools_no_subcommands(runner):
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            main, ["explore", "samtools", "--no-subcommands", "--out-dir", tempdir,],
        )
        cli_worked(result)
        # Since we aren't looking at subcommands, there should be one file for each format
        assert len(list(Path(tempdir).iterdir())) >= 3


def test_grammar(runner):
    result = runner.invoke(main, ["grammar"])
    assert result.exit_code == 0
    assert len(result.output) > 20



import inspect
import tempfile
from io import IOBase, StringIO, TextIOBase
from pathlib import Path
from typing import Generator

from cwl_utils.parser_v1_1 import (
    CommandInputParameter,
    CommandLineBinding,
    CommandLineTool,
)
from dataclasses import dataclass

from acclimatise import cli_types
from acclimatise.converter import WrapperGenerator
from acclimatise.model import Command
from acclimatise.yaml import yaml


def with_default_none(func, *args, **kwargs):
    """
    Calls a function, and inserts None for all arguments you didn't provide
    """
    spec = inspect.getfullargspec(func)
    defaults = {arg: None for arg in spec.args}
    defaults.pop("self", None)
    return func(*args, **{**defaults, **kwargs})


@dataclass
class CwlGenerator(WrapperGenerator):
    case = "snake"

    @classmethod
    def format(cls) -> str:
        return "cwl"

    @staticmethod
    def snake_case(words: list):
        return "_".join([word.lower() for word in words])

    @staticmethod
    def to_cwl_type(typ: cli_types.CliType):
        if isinstance(typ, cli_types.CliFile):
            return "File"
        elif isinstance(typ, cli_types.CliDir):
            return "Directory"
        elif isinstance(typ, cli_types.CliString):
            return "string"
        elif isinstance(typ, cli_types.CliFloat):
            return "double"
        elif isinstance(typ, cli_types.CliInteger):
            return "long"
        elif isinstance(typ, cli_types.CliBoolean):
            return "boolean"
        elif isinstance(typ, cli_types.CliEnum):
            return "string"
        elif isinstance(typ, cli_types.CliList):
            return CwlGenerator.to_cwl_type(typ.value) + "[]"
        elif isinstance(typ, cli_types.CliTuple):
            return [CwlGenerator.to_cwl_type(subtype) for subtype in set(typ.values)]
        else:
            raise Exception(f"Invalid type {typ}!")

    def command_to_tool(self, cmd: Command) -> CommandLineTool:
        """
        Outputs the CWL wrapper to the provided file
        """
        tool = with_default_none(
            CommandLineTool,
            id=cmd.as_filename + ".cwl",
            baseCommand=list(cmd.command),
            cwlVersion="v1.1",
            inputs=[],
            outputs=[],
        )

        if not self.ignore_positionals:
            for pos in cmd.positional:
                tool.inputs.append(
                    with_default_none(
                        CommandInputParameter,
                        id=self.choose_variable_name(pos),
                        type=self.to_cwl_type(pos.get_type()),
                        inputBinding=with_default_none(
                            CommandLineBinding, position=pos.position
                        ),
                        doc=pos.description,
                    )
                )

        for flag in cmd.named:
            tool.inputs.append(
                with_default_none(
                    CommandInputParameter,
                    id=self.choose_variable_name(flag),
                    type=self.to_cwl_type(flag.get_type()),
                    inputBinding=with_default_none(
                        CommandLineBinding, prefix=flag.longest_synonym
                    ),
                    doc=flag.description,
                )
            )
        return tool

    def generate_wrapper(self, cmd: Command) -> str:
        io = StringIO()
        yaml.dump(self.command_to_tool(cmd).save(), io)
        return io.getvalue()

    def generate_tree(self, cmd: Command, out_dir: Path) -> Generator[Path, None, None]:
        for command in cmd.command_tree():
            path = (out_dir / command.as_filename).with_suffix(".cwl")
            with path.open("w") as fp:
                yaml.dump(self.command_to_tool(command).save(), fp)
            yield path



"""
Test data model methods
"""
from acclimatise.model import EmptyFlagArg, Flag


def test_name_to_words():
    """
    Check that we can get an argument name even if the argument's flag is a symbol
    """
    arg = Flag(
        synonyms=["-@"],
        description="Number of additional threads to use",
        args=EmptyFlagArg(),
    )

    assert list(arg.name_to_words()) == ["commercial", "at"]
